#include <fcntl.h>
#include <pthread.h>
#include <stdlib.h>
#include <glib.h>
#include "xmms/plugin.h"
#include "codec.h"

#define BLOCK_SIZE 4096

struct vorbis_data {
   ogg_sync_state oy;
   ogg_stream_state os;
   ogg_page og;
   ogg_packet op;
   vorbis_info vi;
   vorbis_dsp_state vd;
   vorbis_block vb;
};

static void vorbis_init(void);
static int vorbis_begin_decode(int fd, struct vorbis_data *od);
static int vorbis_check_file(char *filename);
static void vorbis_play(char *filename);
static void vorbis_stop(void);

InputPlugin vorbis_ip =
{
   NULL,
   NULL,
   "OggVorbis Player 0.1",
   vorbis_init,
   NULL,
   NULL,
   vorbis_check_file,
   NULL,
   vorbis_play,
   vorbis_stop,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
};

pthread_t tid;

InputPlugin 
*get_iplugin_info(void)
{
   return &vorbis_ip;
}

static void
vorbis_init(void)
{
}

/* Begin the decode procedure.  This is also used to check whether we're
   dealing with vorbis or not */
static int
vorbis_begin_decode(int fd, struct vorbis_data *od)
{
   int bytes;
   char *buffer;
   
   ogg_sync_init(&od->oy);
   
   buffer = ogg_sync_buffer(&od->oy, BLOCK_SIZE);
   bytes = read(fd, buffer, BLOCK_SIZE);
   ogg_sync_wrote(&od->oy, bytes);
   
   if(ogg_sync_pageout(&od->oy, &od->og) != 1) {
      ogg_sync_clear(&od->oy);
      return -1;
   }
   
   ogg_stream_init(&od->os, ogg_page_serialno(&od->og));
   if(ogg_stream_pagein(&od->os, &od->og) < 0) {
      ogg_stream_clear(&od->os);
      ogg_sync_clear(&od->oy);
      return -1;
   }
   
   if(ogg_stream_packetout(&od->os, &od->op) != 1) {
      ogg_stream_clear(&od->os);
      ogg_sync_clear(&od->oy);
      return -1;
   }
   
   vorbis_info_init(&od->vi);
   
   if(vorbis_info_headerin(&od->vi, &od->op) < 0) {
      vorbis_info_clear(&od->vi);
      ogg_stream_clear(&od->os);
      ogg_sync_clear(&od->oy);
      return -1;
   }
   
   return 0;
}

static int 
vorbis_check_file(char *filename)
{
   int fd;
   struct vorbis_data od;
   
   if((fd = open(filename, O_RDONLY)) < 0)
     return FALSE;
   
   if(vorbis_begin_decode(fd, &od) < 0)
     return FALSE;
   
   vorbis_info_clear(&od.vi);
   ogg_stream_clear(&od.os);
   ogg_sync_clear(&od.oy);
   
   return TRUE;
}

static void
*vorbis_play_loop(void *arg)
{
   int fd = (int)arg;
   int bytes, result, samples, i = 0, s, c, eos = 0;
   double **pcm;
   char *buffer;
   struct vorbis_data od;
   gshort *pcmout;
   
   if(vorbis_begin_decode(fd, &od) < 0)
     return NULL;
   
   while(i < 2) {
      while(i < 2 && ogg_sync_pageout(&od.oy, &od.og) == 1) {
	 ogg_stream_pagein(&od.os, &od.og);

	 while(i < 2) {
	    if((result = ogg_stream_packetout(&od.os, &od.op)) < 0) {
	       ogg_sync_clear(&od.oy);
   	       ogg_stream_clear(&od.os);
   	       vorbis_info_clear(&od.vi);
	       return NULL;
	    }
	    if(!result)
	      break;
	    vorbis_info_headerin(&od.vi, &od.op);
	    i++;
	 }
      }
      
      buffer = ogg_sync_buffer(&od.oy, BLOCK_SIZE);
      bytes = read(fd, buffer, BLOCK_SIZE);
      ogg_sync_wrote(&od.oy, bytes);
   }
   
   vorbis_synthesis_init(&od.vd, &od.vi);
   vorbis_block_init(&od.vd, &od.vb);
   
   while(!eos) {
      while(!eos) {
	 if((result = ogg_sync_pageout(&od.oy, &od.og)) < 0) {
	    ogg_sync_clear(&od.oy);
   	    ogg_stream_clear(&od.os);
   
   	    vorbis_info_clear(&od.vi);
   	    vorbis_synthesis_clear(&od.vd);
   	    vorbis_block_clear(&od.vb);
	    return NULL;
	 }
	 if(!result)
	   break;
	 
	 ogg_stream_pagein(&od.os, &od.og);
	 
	 while((result = ogg_stream_packetout(&od.os, &od.op)) != 0) {
	    if(result < 0) {
	       ogg_sync_clear(&od.oy);
   	       ogg_stream_clear(&od.os);
   
   	       vorbis_info_clear(&od.vi);
   	       vorbis_synthesis_clear(&od.vd);
   	       vorbis_block_clear(&od.vb);
	       return NULL;
	    }
	    
	    vorbis_synthesis(&od.vb, &od.op);
	    vorbis_synthesis_blockin(&od.vd, &od.vb);
	    if((samples = vorbis_synthesis_pcmout(&od.vd, &pcm)) > 0) {
	       pcmout = malloc(samples * od.vi.channels * sizeof(gshort));
	       for(s = 0; s < samples; s++)
		 for(c = 0; c < od.vi.channels; c++)
		   pcmout[(s * od.vi.channels + c) * sizeof(gshort)] = pcm[c][s] * 32767;
	       
	       vorbis_ip.add_vis_pcm(vorbis_ip.output->written_time(), FMT_S16_LE, od.vi.channels, samples * od.vi.channels * sizeof(gshort), pcmout);
	       vorbis_ip.add_vis_plugin_pcm(vorbis_ip.output->written_time(), FMT_S16_LE, od.vi.channels, samples * od.vi.channels * sizeof(gshort), pcmout);
	       vorbis_ip.output->write_audio(pcmout, samples * od.vi.channels * sizeof(gshort));
	       
	       free(pcmout);
	       
	       vorbis_synthesis_read(&od.vd, samples);
	    }
	 }
	 if(ogg_page_eos(&od.og))
	   eos = 1;
      }
      
      if(!eos) {
	 buffer = ogg_sync_buffer(&od.oy, BLOCK_SIZE);
	 bytes = read(fd, buffer, BLOCK_SIZE);
	 if(!bytes)
	   eos = 1;
	 else
	   ogg_sync_wrote(&od.oy, bytes);
      }
   }
   
   ogg_sync_clear(&od.oy);
   ogg_stream_clear(&od.os);
   
   vorbis_info_clear(&od.vi);
   vorbis_synthesis_clear(&od.vd);
   vorbis_block_clear(&od.vb);
   
   return NULL;
}

static void
vorbis_play(char *filename)
{
   int fd;
   
   if((fd = open(filename, O_RDONLY)) < 0)
     return;
   
   pthread_create(&tid, NULL, vorbis_play_loop, (void *)fd);
}
 
static void
vorbis_stop(void)
{
   pthread_join(tid, NULL);
   
   vorbis_ip.output->close_audio();
}
